﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】
using System.Runtime.InteropServices; //【Marshal.Copy】
using System.Threading;               //【Threadを使う為に必要】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択
using Microsoft.Speech.AudioFormat;   //【参照設定】C:\Program Files\Microsoft SDKs\Kinect\v1.0\Assemblies\Microsoft.Kinect.dll
using Microsoft.Speech.Recognition;   //【参照設定】C:\Program Files\Microsoft SDKs\Speech\v11.0\Assembly\Microsoft.Speech.dll


namespace KinectSpeech
{
    public partial class Form1 : Form
    {
        //【KINECT関連】--------------------------------------------------------------------
        private KinectSensor KS_KinectSensor;   //【KINECTセンサ】

        //【音声認識関連】
        private KinectAudioSource KS_Source;    //【KINECTオーディオ入力】
        private RecognizerInfo KS_RI;           //【音声認識情報】
        private SpeechRecognitionEngine KS_SRE; //【音声認識エンジン】
        private Choices KS_SelectiveWords;      //【音声認識候補の語】
        private GrammarBuilder KS_GB;           //【文法生成】
        private Grammar KS_G;                   //【文法】
        private string[] KS_Choice = { "red", "yellow", "green", "blue", "start", "stop", "right", "left", "up", "down", "exit",
                                       "measure", "sensor", "audio", "sound", "hello", "Japan" };
        private string KS_RecognizedWord;
        private string KS_PastWord = "";
        private float KS_Confidence;

        //【カラー画像キャプチャー関係】
        private ColorImageFrame KS_ColorFrame;  //【KINECTのColorImageFrame】
        private byte[] KS_ColorImageDataXRGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private int KS_ColorFrameWidth;         //【KINECTのColorFrameの横方向画素数】
        private int KS_ColorFrameHeight;        //【KINECTのColorFrameの縦方向画素数】
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private bool KS_flagFlip = true;        //【左右反転フラグ】

        public Form1()
        {
            InitializeComponent();
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            if (KS_KinectSensor == null)
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //【KINECTの設定】ColorStreamを許可*******************************************************************
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】********************************************
            KS_KinectSensor.ColorFrameReady += ColorImageReady; //【準備完了でColorImageReadyがcallされる】
            KS_ColorImageDataXRGB = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength];      //【カラー画像データ用】
            KS_ColorFrameWidth = KS_KinectSensor.ColorStream.FrameWidth;   //【ColorImageの横方向画素数】
            KS_ColorFrameHeight = KS_KinectSensor.ColorStream.FrameHeight; //【ColorImageの縦方向画素数】
            KS_KinectSensor.ElevationAngle = 0; //【KINECTセンサの上下角を0°に設定】設定範囲±27°
            KS_bmp1 = new Bitmap(KS_ColorFrameWidth, KS_ColorFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ColorFrameWidth, KS_ColorFrameHeight); //【ColorImage用枠】

            KS_Source = KS_KinectSensor.AudioSource;
            KS_Source.EchoCancellationMode = EchoCancellationMode.None; //【オーディオ・エコーキャンセラーを使わない】
            KS_Source.AutomaticGainControlEnabled = false; //【音声認識の為に自動ゲイン・コントロールをオフにする】
            KS_Source.SoundSourceAngleChanged +=new EventHandler<SoundSourceAngleChangedEventArgs>(SoundSourceAngleChanged);
            KS_Source.BeamAngleChanged +=new EventHandler<BeamAngleChangedEventArgs>(BeamAngleChanged);

            KS_RI = GetKinectRecognizer();
            if (KS_RI == null)
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                    "KINECT音声認識が見つかりません。KINECTの接続をご確認ください。",
                    "KINECT for Windows Error Message",
                    MessageBoxButtons.OK
                );
                return;
            }
            label2.Text = KS_RI.Name; //【取得した音声認識名称を表示する】

            int wait = 4; //デバイスが正常に動作するように初期化後４秒間待つ必要がある。
            while (wait > 0)
            {
                label5.Text = "デバイスが音声認識の準備を完了するまであと" + wait.ToString() + "秒です。";
                wait--;
                Thread.Sleep(1000);
            }

            KS_SRE = new SpeechRecognitionEngine(KS_RI.Id);
            KS_SelectiveWords = new Choices();
            for (int i = 0; i < KS_Choice.Length; i++) { KS_SelectiveWords.Add(KS_Choice[i]); }
            
            KS_GB = new GrammarBuilder();
            KS_GB.Culture = KS_RI.Culture; // Specify the culture to match the recognizer in case we are running in a different culture.                                 
            KS_GB.Append(KS_SelectiveWords); 
            KS_G = new Grammar(KS_GB); //文法のインスタンスを作成する
            KS_SRE.LoadGrammar(KS_G);  //音声認識に文法インスタンスをロードする
            KS_SRE.SpeechRecognized += SreSpeechRecognized; //音声認識処理実行イベントのcall先をSreSpeechRecognizedに設定する
            Stream s=KS_Source.Start();
            KS_SRE.SetInputToAudioStream(s, new SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000, 16, 1, 32000, 2, null));
            label5.Text = "音声認識実行中。登録語を発音してみてください。";
            KS_SRE.RecognizeAsync(RecognizeMode.Multiple);
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (KS_ColorFrame = e.OpenColorImageFrame())
            {
                if (KS_ColorFrame != null)
                {
                    KS_ColorFrame.CopyPixelDataTo(KS_ColorImageDataXRGB); //【KINECTのColor画像を配列PixelDataにコピー】
                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataXRGB, 0, KS_data1.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;            //【pictureBox1に画像表示】
                    //             Application.DoEvents();
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }

                label5.Text = "【音声認識された語】" + KS_RecognizedWord + "  【確からしさ】" + KS_Confidence.ToString();
                switch (KS_RecognizedWord)
                {
                    case "red":
                       pictureBox3.Image=System.Drawing.Image.FromFile("../../image/Red.jpg"); break;
                    case "yellow":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Yellow.jpg"); break;
                    case "green":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Green.jpg"); break;
                    case "blue":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Blue.jpg"); break;
                    case "start":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Start.jpg"); break;
                    case "stop":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Stop.jpg"); break;
                    case "right":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Right.jpg"); break;
                    case "left":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Left.jpg"); break;
                    case "up":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Up.jpg");
                        if (KS_PastWord != KS_RecognizedWord) { KS_KinectSensor.ElevationAngle = 27; textBox1.Text = "27"; }
                        break;
                    case "down":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Down.jpg");
                        if (KS_PastWord != KS_RecognizedWord) { KS_KinectSensor.ElevationAngle = -27; textBox1.Text = "-27"; }
                        break;
                    case "exit":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Exit.jpg"); break;
                    case "measure":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Measure.jpg"); break;
                    case "sensor":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Sensor.jpg"); break;
                    case "audio":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Audio.jpg"); break;
                    case "sound":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Sound.jpg"); break;
                    case "hello":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Hello.jpg"); break;
                    case "Japan":
                        pictureBox3.Image = System.Drawing.Image.FromFile("../../image/Japan.jpg"); break;
                }
                KS_PastWord = KS_RecognizedWord;
            }
        }

        //***************************************************************
        //【終了要求ボタン】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
                KS_SRE.RecognizeAsyncStop();
                KS_Source.Stop();
            }
            this.Close();              //【Formを閉じる】
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【GetKinectRecognizer】
        //***************************************************************
        private static RecognizerInfo GetKinectRecognizer()
        {
            Func<RecognizerInfo, bool> matchingFunc = r =>
            {
                string value;
                r.AdditionalInfo.TryGetValue("Kinect", out value);
                return "True".Equals(value, StringComparison.InvariantCultureIgnoreCase)
                    && "en-US".Equals(r.Culture.Name, StringComparison.InvariantCultureIgnoreCase);
            };
            return SpeechRecognitionEngine.InstalledRecognizers().Where(matchingFunc).FirstOrDefault();
        }

        //***************************************************************
        //【SreSpeechRecognized】 
        //***************************************************************
        private void SreSpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            KS_Confidence = e.Result.Confidence;
            if (e.Result.Confidence >= 0.7)
            {
                KS_RecognizedWord = e.Result.Text;
            }
        }

        //***************************************************************
        //【SreSpeechRecognitionRejected】static 
        //***************************************************************
        private void SreSpeechRecognitionRejected(object sender, SpeechRecognitionRejectedEventArgs e)
        {
            Console.WriteLine("\nSpeech Rejected");
            if (e.Result != null)
            {
                DumpRecordedAudio(e.Result.Audio);
            }
        }

        //***************************************************************
        //【SreSpeechHypothesized】static 
        //***************************************************************
        private void SreSpeechHypothesized(object sender, SpeechHypothesizedEventArgs e)
        {
            Console.Write("\rSpeech Hypothesized: \t{0}", e.Result.Text);
        }

        //***************************************************************
        //【DumpRecordedAudio】
        //***************************************************************
        private static void DumpRecordedAudio(RecognizedAudio audio)
        {
            if (audio == null) { return; }

            int fileId = 0;
            string filename;
            while (File.Exists((filename = "RetainedAudio_" + fileId + ".wav"))) { fileId++; } //【】

            Console.WriteLine("\nWriting file: {0}", filename);
            using (var file = new FileStream(filename, System.IO.FileMode.CreateNew))
            {
                audio.WriteToWaveStream(file);
            }
        }

        //***************************************************************
        //【SoundSourceAngleChangedの取得イベント】
        //***************************************************************
        private void SoundSourceAngleChanged(object sender, SoundSourceAngleChangedEventArgs e)
        {
            label3.Text = "【音源の角度】" + e.Angle.ToString() + "[度]";
        }

        //***************************************************************
        //【BeamAngleChangedの取得イベント】
        //***************************************************************
        private void BeamAngleChanged(object sender, BeamAngleChangedEventArgs e)
        {
            label4.Text = "【ビームの角度】" + e.Angle.ToString() + "[度]";
        }
    }
}
